"""Prepare params for request in TweetsByIdsRunner."""
from ..http_request import HttpMethod
from ..http_request import RequestDetails

_request_url = 'https://cdn.syndication.twimg.com/tweet'


def get_request_details_for_base_tweet_info(tweet_id: str) -> RequestDetails:
    """Prepare params for request in TweetsByIdsRunner."""
    return RequestDetails(
        http_method=HttpMethod.GET,
        url=_request_url,
        headers=dict(),
        params=dict({'id': tweet_id}),
        timeout=10
    )



"""Domain Language enum class."""

import enum


class Language(enum.Enum):
    """Domain Language enum class."""

    def __new__(cls, *args, **kwargs):
        """Class __new__ method."""
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, short_value):
        """Class constructor method."""
        self.short_value = short_value

    ENGLISH = 'en'
    ARABIC = 'ar'
    BASQUE = 'eu'
    BENGALI = 'bn'
    BULGARIAN = 'bg'
    TRADITIONAL_CHINESE = 'zh-tw'
    SIMPLIFIED_CHINESE = 'zh-cn'
    CROATIAN = 'hr'
    CZECH = 'cs'
    DANISH = 'da'
    FINNISH = 'fi'
    FRENCH = 'fr'
    GREEK = 'el'
    GUJARATI = 'gu'
    HEBREW = 'iw'
    HINDI = 'hi'
    SPANISH = 'es'
    INDONESIAN = 'in'
    JAPANESE = 'ja'
    CANADIAN = 'kn'
    CATALAN = 'ca'
    KOREAN = 'ko'
    MARATHI = 'mr'
    DUTCH = 'nl'
    GERMAN = 'de'
    NORWEGIAN = 'no'
    PERSIAN = 'fa'
    POLISH = 'pl'
    PORTUGUESE = 'pt'
    RUSSIAN = 'ru'
    ROMANIAN = 'ro'
    SERBIAN = 'sr'
    SLOVAK = 'sk'
    SWEDISH = 'sv'
    THAI = 'th'
    TAMIL = 'ta'
    TURKISH = 'tr'
    UKRAINIAN = 'uk'
    URDU = 'ur'
    HUNGARIAN = 'hu'
    VIETNAMESE = 'vi'
    ITALIAN = 'it'



"""Utils to parse data from web response."""

from abc import abstractmethod
from typing import List, Optional

from ...model.tweet import Tweet


class TweetParser:
    """Utils class to parse data from web response."""

    @abstractmethod
    def parse_tweets(self, response_text: str) -> List[Tweet]:
        """Method to extract tweets from web response."""

    @abstractmethod
    def parse_cursor(self, response_content: str) -> Optional[str]:
        """Method to extract next cursor to scrap request from web response."""



"""Domain TweetsByIdsTask class."""
from dataclasses import dataclass
from typing import List


@dataclass(frozen=True)
class TweetsByIdsTask:
    """Domain TweetsByIdsTask class."""

    tweet_ids: List[str]

    def __init__(
            self,
            tweet_ids: List[str]
    ):
        """Class constructor."""
        object.__setattr__(self, 'tweet_ids', tweet_ids)
        return



"""Class with request details."""

from dataclasses import dataclass
from typing import Dict

from .http_method import HttpMethod


@dataclass
class RequestDetails:
    """Class with request details. Specify all http request details."""

    http_method: HttpMethod
    url: str
    headers: Dict[str, str]
    params: Dict[str, str]
    timeout: int



"""Methods to read users from files."""

from typing import List

import pandas as pd

from ..mapper.user_dict_mapper import create_user_from_flat_dict
from ..mapper.user_json_mapper import create_user_from_json
from ..model import User


def read_users_from_csv_file(file_path: str) -> List[User]:
    """Method to read tweets from csv file."""
    df = pd.read_csv(file_path, dtype={
        'pinned_tweet_ids_str': str,
        'profile_banner_url': str,
        'location': str,
        'description': str
    })
    df.pinned_tweet_ids_str.fillna('', inplace=True)
    df.profile_banner_url.fillna('', inplace=True)
    df.location.fillna('', inplace=True)
    df.description.fillna('', inplace=True)
    return [create_user_from_flat_dict(row) for _, row in df.iterrows()]


def read_users_from_json_lines_file(file_path: str) -> List[User]:
    """Method to read tweets from csv file."""
    file = open(file_path, 'r')
    return [create_user_from_json(line) for line in file.readlines()]



"""Parser of JSON string to User."""
import json

from arrow import Arrow
from dateutil import parser

from ..model import User


def parse_user(response_content: str) -> User:
    """Parser of JSON string to User."""
    user_json = json.loads(response_content)['data']['user']
    legacy_user_json = user_json['legacy']
    return User(
        created_at=Arrow.fromdatetime(parser.parse(legacy_user_json['created_at'])),
        id_str=user_json['id'],
        rest_id_str=user_json['rest_id'],
        default_profile=legacy_user_json['default_profile'],
        default_profile_image=legacy_user_json['default_profile_image'],
        description=legacy_user_json['description'],
        favourites_count=legacy_user_json['favourites_count'],
        followers_count=legacy_user_json['favourites_count'],
        friends_count=legacy_user_json['friends_count'],
        has_custom_timelines=legacy_user_json['has_custom_timelines'],
        listed_count=legacy_user_json['listed_count'],
        location=legacy_user_json['location'],
        media_count=legacy_user_json['media_count'],
        name=legacy_user_json['name'],
        pinned_tweet_ids_str=legacy_user_json['pinned_tweet_ids_str'],
        profile_banner_url=legacy_user_json['profile_banner_url'] if 'profile_banner_url' in legacy_user_json else '',
        profile_image_url_https=legacy_user_json['profile_image_url_https'],
        protected=legacy_user_json['protected'],
        screen_name=legacy_user_json['screen_name'],
        statuses_count=legacy_user_json['statuses_count'],
        verified=legacy_user_json['verified']
    )



from .http_method import HttpMethod
from .request_details import RequestDetails
from .request_response import RequestResponse
from .requests.requests_web_client import RequestsWebClient, RequestsWebClientProxyConfig
from .web_client import WebClient



"""Domain TweetsByIdsContext class."""

from dataclasses import dataclass
from typing import Optional


@dataclass
class GetUsersContext:
    """Domain TweetsByIdsContext class."""

    guest_auth_token: Optional[str] = None
    scrapped_count: int = 0

    def add_one_scrapped_user(self):
        """Method raise counter of all scrapped tweets."""
        self.scrapped_count += 1



"""Domain Tweet class."""

from dataclasses import dataclass
from typing import List

from arrow import Arrow


@dataclass
class Tweet:
    """Domain Tweet class."""

    created_at: Arrow
    id_str: str
    conversation_id_str: str
    full_text: str
    lang: str
    favorited: bool
    retweeted: bool
    retweet_count: int
    favorite_count: int
    reply_count: int
    quote_count: int
    quoted_status_id_str: str
    quoted_status_short_url: str
    quoted_status_expand_url: str
    user_id_str: str
    user_name: str
    user_full_name: str
    user_verified: bool
    in_reply_to_status_id_str: str
    in_reply_to_user_id_str: str
    media_url: str
    hashtags: List[str]
    mentions: List[str]
    urls: List[str]



"""Configuration of proxy to RequestsWebClient."""
from dataclasses import dataclass


@dataclass
class RequestsWebClientProxyConfig:
    """Configuration class of proxy to RequestsWebClient."""

    http_proxy: str
    https_proxy: str



"""Methods to export_data users."""
from typing import List

from .util import clear_file
from ..model import User
from ..user_output import CsvUserOutput, JsonLineFileUserOutput


def export_users_to_csv(users: List[User], filename: str):
    """Method to export_data users to csv."""
    clear_file(filename)
    CsvUserOutput(filename).export_users(users)
    return


def export_users_to_json_lines(users: List[User], filename: str):
    """Method to export_data users to json lines."""
    clear_file(filename)
    JsonLineFileUserOutput(filename).export_users(users)
    return



"""Utils to prepare data."""
from typing import Dict

from ..http_request import HttpMethod
from ..http_request import RequestDetails
from ..search_runner.search_run_context import SearchRunContext
from ..search_runner.search_tweets_task import SearchTweetsTask

_bearer_token = 'Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4pu' \
                'Ts%3D1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA'
_default_tweets_count_in_batch = 100
_search_tweets_url = 'https://api.twitter.com/2/search/adaptive.json'
_timeout = 10


def _scrap_tweets_get_params(
        search_run_context: SearchRunContext,
        search_tweets_task: SearchTweetsTask
) -> Dict[str, str]:
    """Method to prepare requests params to get tweets."""
    count_in_batch = _default_tweets_count_in_batch \
        if search_tweets_task.tweets_limit is None \
        else min(_default_tweets_count_in_batch,
                 search_tweets_task.tweets_limit - search_run_context.all_download_tweets_count)
    return dict([
        ('include_can_media_tag', '1'),
        ('include_ext_alt_text', 'true'),
        ('include_quote_count', 'true'),
        ('include_reply_count', '1'),
        ('tweet_mode', 'extended'),
        ('include_entities', 'true'),
        ('include_user_entities', 'true'),
        ('include_ext_media_availability', 'true'),
        ('send_error_codes', 'true'),
        ('simple_quoted_tweet', 'true'),
        ('count', count_in_batch),
        ('cursor', search_run_context.scroll_token),
        ('spelling_corrections', '1'),
        ('ext', 'mediaStats%2ChighlightedLabel'),
        ('tweet_search_mode', 'live'),
        ('f', 'tweets'),  # if not config.Popular_tweets
        ('q', search_tweets_task.get_full_search_query())
    ])


def _scrap_tweets_get_headers(search_run_context: SearchRunContext) -> Dict[str, str]:
    """Method to prepare requests headers to get tweets."""
    return dict([
        ("authorization", _bearer_token),
        ("x-guest-token", search_run_context.guest_auth_token)
    ])


def get_search_tweet_request_details(
        search_run_context: SearchRunContext,
        search_tweets_task: SearchTweetsTask
) -> RequestDetails:
    """Returns the RequestDetails for simple request for tweets."""
    return RequestDetails(
        HttpMethod.GET,
        url=_search_tweets_url,
        headers=_scrap_tweets_get_headers(search_run_context),
        params=_scrap_tweets_get_params(search_run_context, search_tweets_task),
        timeout=_timeout
    )



"""TweetOutput which print each n-th tweet."""

from typing import List

from .tweet_output import TweetOutput
from ..model.tweet import Tweet


class PrintEveryNTweetOutput(TweetOutput):
    """TweetOutput which print each n-th tweet."""

    each_n: int
    _counter: int = 0

    def __init__(self, each_n: int):
        """Create instance of PrintEveryNTweetOutput."""
        self.each_n = each_n

    def export_tweets(self, tweets: List[Tweet]):
        """Method print first each n tweet in console with all scrapped count."""
        for it in tweets:
            self._counter += 1
            if self._counter % self.each_n == 0:
                print(self._counter, it)
        return



"""Tweet - dict mapper."""
from copy import copy
from typing import Dict

from arrow import get as arrow_get

from .util import simple_string_list_to_string, string_to_simple_string_list
from ..model import Tweet


def tweet_to_flat_dict(self):
    """Method to prepare flat dict of tweet. Used in CSV serialization."""
    dictionary = dict(self.__dict__)
    dictionary['hashtags'] = simple_string_list_to_string(dictionary['hashtags'])
    dictionary['mentions'] = simple_string_list_to_string(dictionary['mentions'])
    dictionary['urls'] = simple_string_list_to_string(dictionary['urls'])
    return dictionary


def create_tweet_from_dict(dictionary: Dict[str, any]):
    """Method to create Tweet from dictionary."""
    return Tweet(
        arrow_get(dictionary['created_at']),
        str(dictionary['id_str']),
        str(dictionary['conversation_id_str']),
        dictionary['full_text'],
        dictionary['lang'],
        dictionary['favorited'],
        dictionary['retweeted'],
        dictionary['retweet_count'],
        dictionary['favorite_count'],
        dictionary['reply_count'],
        dictionary['quote_count'],
        dictionary['quoted_status_id_str'],
        dictionary['quoted_status_short_url'],
        dictionary['quoted_status_expand_url'],
        str(dictionary['user_id_str']),
        dictionary['user_name'],
        dictionary['user_full_name'],
        dictionary['user_verified'],
        str(dictionary['in_reply_to_status_id_str']),
        str(dictionary['in_reply_to_user_id_str']),
        dictionary['media_url'],
        dictionary['hashtags'],
        dictionary['mentions'],
        dictionary['urls']
    )


def create_tweet_from_flat_dict(dictionary: Dict[str, any]) -> Tweet:
    """Method to create Tweet from flat dictionary."""
    dictionary = copy(dictionary)
    dictionary['hashtags'] = string_to_simple_string_list(dictionary['hashtags'])
    dictionary['mentions'] = string_to_simple_string_list(dictionary['mentions'])
    dictionary['urls'] = string_to_simple_string_list(dictionary['urls'])
    return create_tweet_from_dict(dictionary)



"""Runner for get tweets by ids."""
import json
from dataclasses import dataclass
from typing import List, Optional

from arrow import Arrow, get as arrow_get

from stweet.http_request.requests.requests_web_client import RequestsWebClient
from .tweets_by_ids_context import TweetsByIdsContext
from .tweets_by_ids_request_details_builder import get_request_details_for_base_tweet_info
from .tweets_by_ids_result import TweetsByIdsResult
from .tweets_by_ids_task import TweetsByIdsTask
from ..auth import AuthTokenProviderFactory, SimpleAuthTokenProviderFactory
from ..http_request import WebClient
from ..model import Tweet
from ..search_runner import TweetSearchRunner
from ..search_runner.parse import BaseTweetParser
from ..search_runner.parse import TweetParser
from ..search_runner.search_run_context import SearchRunContext
from ..search_runner.search_tweets_task import SearchTweetsTask
from ..tweet_output import TweetOutput, CollectorTweetOutput


@dataclass
class _TweetByIdBaseInfo:
    id: str
    username: str
    tweet_content: str
    created_at: Arrow


class TweetsByIdsRunner:
    """Runner class to process task to search tweets."""

    tweets_by_ids_context: TweetsByIdsContext
    tweets_by_ids_task: TweetsByIdsTask
    tweet_outputs: List[TweetOutput]
    web_client: WebClient
    tweet_parser: TweetParser
    auth_token_provider_factory: AuthTokenProviderFactory

    def __init__(
            self,
            tweets_by_ids_task: TweetsByIdsTask,
            tweet_outputs: List[TweetOutput],
            tweets_by_ids_context: Optional[TweetsByIdsContext] = None,
            web_client: WebClient = RequestsWebClient(),
            tweet_parser: TweetParser = BaseTweetParser(),
            auth_token_provider_factory: AuthTokenProviderFactory = SimpleAuthTokenProviderFactory()
    ):
        """Constructor to create object."""
        self.tweets_by_ids_context = TweetsByIdsContext() if tweets_by_ids_context is None else tweets_by_ids_context
        self.tweets_by_ids_task = tweets_by_ids_task
        self.tweet_outputs = tweet_outputs
        self.web_client = web_client
        self.tweet_parser = tweet_parser
        self.auth_token_provider_factory = auth_token_provider_factory
        return

    def run(self) -> TweetsByIdsResult:
        """Main search_runner method."""
        self._prepare_token()
        tweet_ids_not_scrapped = []
        for tweet_id_to_scrap in self.tweets_by_ids_task.tweet_ids:
            tweet_base_info = self._get_base_tweet_info(tweet_id_to_scrap)
            if tweet_base_info is None:
                tweet_ids_not_scrapped.append(tweet_id_to_scrap)
            else:
                tweet = self._scrap_full_tweet(tweet_base_info)
                if tweet is None:
                    tweet_ids_not_scrapped.append(tweet_id_to_scrap)
                else:
                    self.tweets_by_ids_context.add_downloaded_tweets_count(1)
                    self._process_new_tweets_to_output([tweet])
        return TweetsByIdsResult(self.tweets_by_ids_context.all_download_tweets_count, tweet_ids_not_scrapped)

    def _get_base_tweet_info(self, tweet_id: str) -> Optional[_TweetByIdBaseInfo]:
        request_params = get_request_details_for_base_tweet_info(tweet_id)
        request_result = self.web_client.run_request(request_params)
        return self._get_base_tweet_info_from_text_response(tweet_id, request_result.text) \
            if request_result.is_success() \
            else None

    @staticmethod
    def _get_base_tweet_info_from_text_response(tweet_id: str, response_text: str) -> Optional[_TweetByIdBaseInfo]:
        parsed_json = json.loads(response_text)
        return _TweetByIdBaseInfo(
            tweet_id,
            parsed_json['user']['screen_name'],
            parsed_json['text'],
            arrow_get(parsed_json['created_at'])
        )

    def _scrap_full_tweet(self, tweet_base_info: _TweetByIdBaseInfo) -> Optional[Tweet]:
        search_tweets_task = SearchTweetsTask(
            from_username=tweet_base_info.username,
            since=tweet_base_info.created_at.shift(minutes=-1),
            until=tweet_base_info.created_at.shift(minutes=1)
        )
        tweets_collector = CollectorTweetOutput()
        search_context = SearchRunContext(guest_auth_token=self.tweets_by_ids_context.guest_auth_token)
        TweetSearchRunner(
            search_tweets_task=search_tweets_task,
            tweet_outputs=[tweets_collector],
            web_client=self.web_client,
            search_run_context=search_context,
            tweet_parser=self.tweet_parser,
            auth_token_provider_factory=self.auth_token_provider_factory
        ).run()
        self.tweets_by_ids_context.guest_auth_token = search_context.guest_auth_token
        filtered_list = [it for it in tweets_collector.get_scrapped_tweets() if it.id_str == tweet_base_info.id]
        return filtered_list[0] if len(filtered_list) > 0 else None

    def _refresh_token(self):
        token_provider = self.auth_token_provider_factory.create(self.web_client)
        self.tweets_by_ids_context.guest_auth_token = token_provider.get_new_token()
        return

    def _prepare_token(self):
        if self.tweets_by_ids_context.guest_auth_token is None:
            self._refresh_token()
        return

    def _process_new_tweets_to_output(self, new_tweets: List[Tweet]):
        for tweet_output in self.tweet_outputs:
            tweet_output.export_tweets(new_tweets)
        return



"""Class with result of TweetSearchRunner task."""

from dataclasses import dataclass
from typing import List


@dataclass
class TweetsByIdsResult:
    """Class with result of TweetSearchRunner task."""

    downloaded_count: int
    tweet_ids_not_scrapped: List[str]



from .export_data import export_users_to_csv, export_users_to_json_lines, export_tweets_to_csv, \
    export_tweets_to_json_lines
from .get_user_runner import GetUsersTask, GetUsersRunner, GetUsersResult
from .http_request import WebClient, RequestsWebClient, RequestsWebClientProxyConfig
from .import_data import read_tweets_from_csv_file, read_tweets_from_json_lines_file, read_users_from_csv_file, \
    read_users_from_json_lines_file
from .model import Language, Tweet, User
from .search_runner import SearchTweetsResult, TweetSearchRunner, SearchTweetsTask, RepliesFilter
from .tweet_output import CollectorTweetOutput, CsvTweetOutput, JsonLineFileTweetOutput, \
    PrintEveryNTweetOutput, PrintTweetOutput, TweetOutput, PrintFirstInRequestTweetOutput
from .tweets_by_ids_runner import TweetsByIdsResult, TweetsByIdsTask, TweetsByIdsRunner
from .user_output import UserOutput, PrintUserOutput, CollectorUserOutput, CsvUserOutput, JsonLineFileUserOutput, \
    PrintEveryNUserOutput



"""UserOutput which saves user data by request received users batch to CSV file."""

import os
from typing import List

import pandas as pd

from .user_output import UserOutput
from ..mapper.user_dict_mapper import user_to_flat_dict
from ..model import User

_fields_list = [
    'created_at', 'id_str', 'rest_id_str', 'default_profile', 'default_profile_image', 'description',
    'favourites_count', 'followers_count', 'friends_count', 'has_custom_timelines', 'listed_count', 'location',
    'media_count', 'name', 'pinned_tweet_ids_str', 'profile_banner_url', 'profile_banner_url',
    'profile_image_url_https', 'protected', 'screen_name', 'statuses_count', 'verified'
]


class CsvUserOutput(UserOutput):
    """CsvUserOutput saves user data by request received users batch to CSV file."""

    file_location: str
    add_header_on_start: bool

    def __init__(self, file_location: str, add_header_on_start: bool = True):
        """Create instance of CsvUserOutput."""
        self.file_location = file_location
        self.add_header_on_start = add_header_on_start

    def export_users(self, users: List[User]):
        """Export users to CSV."""
        dict_list = [user_to_flat_dict(it) for it in users]
        df = pd.DataFrame(dict_list, columns=_fields_list)
        df.to_csv(
            path_or_buf=self.file_location,
            mode='a',
            header=self._header_to_add(),
            index=False
        )
        return

    def _header_to_add(self):
        return (not self._file_exists()) or self._file_is_empty()

    def _file_exists(self) -> bool:
        return os.path.isfile(self.file_location)

    def _file_is_empty(self) -> bool:
        return os.stat(self.file_location).st_size == 0



"""UserOutput which saves user data in memory, after request can return all saved users."""
from typing import List

from .user_output import UserOutput
from ..model import User


class CollectorUserOutput(UserOutput):
    """UserOutput which print first user from request batch."""

    _users: List[User]

    def __init__(self):
        """Constructor of class, initialize users as empty list."""
        self._users = []

    def export_users(self, users: List[User]):
        """Method print first user in console, when users list is empty there message will be printed."""
        self._users.extend(users)
        return

    def get_scrapped_users(self) -> List[User]:
        """Method returns all collected users."""
        return self._users



import glob
import os
import string
import unicodedata
import uuid
from typing import List, Callable

import stweet as st

_temp_file_prefix = 'test_temp_file_'


def get_temp_test_file_name(file_extension_without_dot: str) -> str:
    return '{}{}.{}'.format(_temp_file_prefix, _get_uuid_str(), file_extension_without_dot)


def _get_uuid_str() -> str:
    return str(uuid.uuid4()).replace('-', '')


def remove_all_temp_files():
    files_to_remove = glob.glob("{}*".format(_temp_file_prefix))
    for filePath in files_to_remove:
        os.remove(filePath)
    return


def _remove_accented_chars(text) -> str:
    new_text = unicodedata.normalize('NFKD', text).encode('ascii', 'ignore').decode('utf-8', 'ignore')
    return new_text


def to_base_text(value: str) -> str:
    table = str.maketrans(dict.fromkeys(string.punctuation))
    to_return = _remove_accented_chars(value.translate(table).lower())
    return to_return


def get_tweets_to_tweet_output_test(tweet_output: List[st.TweetOutput]):
    phrase = '#koronawirus'
    search_tweets_task = st.SearchTweetsTask(
        all_words=phrase,
        tweets_limit=200
    )
    st.TweetSearchRunner(
        search_tweets_task=search_tweets_task,
        tweet_outputs=tweet_output
    ).run()


def get_users_to_tweet_output_test(user_outputs: List[st.UserOutput]) -> st.GetUsersResult:
    users = ['RealDonaldTrump', 'ProtasiewiczJ', 'donaldtuskEPP', 'RealDonaldTrump', 'ProtasiewiczJ', 'donaldtuskEPP']
    get_users_task = st.GetUsersTask(users)
    return st.GetUsersRunner(
        get_user_task=get_users_task,
        user_outputs=user_outputs
    ).run()


def tweet_list_assert_condition(tweets: List[st.Tweet], condition: Callable[[st.Tweet], bool]):
    for tweet in tweets:
        if not condition(tweet):
            print(f'--- {tweet}')
    assert all([
        condition(tweet)
        for tweet in tweets
    ]) is True


def two_lists_assert_equal(tweets_1: List[any], tweets_2: List[any]):
    assert len(tweets_1) == len(tweets_2)
    for tweet_id in range(len(tweets_1)):
        if tweets_1[tweet_id] != tweets_2[tweet_id]:
            print('-----')
            print(tweets_1[tweet_id])
            print(tweets_2[tweet_id])
            print('--')
    assert tweets_1 == tweets_2



"""Request search_runner class."""
from typing import Optional, Dict

import requests

from .requests_web_client_proxy_config import RequestsWebClientProxyConfig
from ..request_details import RequestDetails
from ..request_response import RequestResponse
from ..web_client import WebClient


class RequestsWebClient(WebClient):
    """Request search_runner class. Implementation based on requests library."""

    proxy: Optional[RequestsWebClientProxyConfig]
    verify: bool

    def __init__(
            self,
            proxy: Optional[RequestsWebClientProxyConfig] = None,
            verify: bool = True
    ):
        """Constructor of RequestsWebClient."""
        self.proxy = proxy
        self.verify = verify

    def run_request(self, params: RequestDetails) -> RequestResponse:
        """Main method to run request using requests package."""
        session = requests.Session()
        response = session.request(
            method=params.http_method.name,
            url=params.url,
            params=params.params,
            headers=params.headers,
            timeout=params.timeout,
            proxies=self._get_proxy(),
            verify=self.verify
        )
        return RequestResponse(response.status_code, response.text)

    def _get_proxy(self) -> Dict[str, str]:
        return None if self.proxy is None else dict({
            'http': self.proxy.http_proxy,
            'https': self.proxy.https_proxy,
        })



"""User - json mapper."""
import json

from arrow import Arrow

from .user_dict_mapper import create_user_from_dict
from ..model import User


def user_to_json(user: User) -> str:
    """Method to prepare json of user. Used in JSON serialization."""
    return json.dumps(user, default=lambda o: str(o) if isinstance(o, Arrow) else o.__dict__)


def create_user_from_json(json_value: str) -> User:
    """Creates user from json string."""
    tweet_dict = json.loads(json_value)
    return create_user_from_dict(tweet_dict)



from typing import Optional

import stweet as st
from stweet.http_request import RequestDetails, RequestResponse


class MockWebClient(st.WebClient):
    status_code: Optional[int]
    text: Optional[str]

    def __init__(self, status_code: Optional[int], text: Optional[str]):
        self.status_code = status_code
        self.text = text

    def run_request(self, params: RequestDetails) -> RequestResponse:
        return RequestResponse(self.status_code, self.text)




"""Mapping util."""
from typing import List

_list_separator = ' , '


def simple_string_list_to_string(values: List[str]) -> str:
    """Method to build string from list."""
    return _list_separator.join(values)


def string_to_simple_string_list(value: str) -> List[str]:
    """Method to build list from specific string."""
    return value.split(_list_separator) if len(value) > 0 else []



"""Class with result of TweetSearchRunner task."""

from dataclasses import dataclass


@dataclass
class SearchTweetsResult:
    """Class with result of TweetSearchRunner task."""

    downloaded_count: int



"""TweetOutput which print first tweet from request batch."""

from typing import List

from .tweet_output import TweetOutput
from ..model.tweet import Tweet


class PrintFirstInRequestTweetOutput(TweetOutput):
    """TweetOutput which print first tweet from request batch."""

    def export_tweets(self, tweets: List[Tweet]):
        """Method print first tweet in console, when tweets list is empty there message will be printed."""
        message = str(tweets[0]) if len(tweets) > 0 else 'PrintFirstInRequestTweetOutput -- no tweets to print'
        print(message)
        return



"""Methods to read tweets from files."""

from typing import List

import pandas as pd

from ..mapper.tweet_dict_mapper import create_tweet_from_flat_dict
from ..mapper.tweet_json_mapper import create_tweet_from_json
from ..model.tweet import Tweet


def read_tweets_from_csv_file(file_path: str) -> List[Tweet]:
    """Method to read tweets from csv file."""
    df = pd.read_csv(file_path, dtype={
        'quoted_status_id_str': str,
        'in_reply_to_status_id_str': str,
        'in_reply_to_user_id_str': str,
        'media_url': str
    })
    if 'media_url' not in df.columns:
        df['media_url'] = ''
    df.quoted_status_id_str.fillna('', inplace=True)
    df.quoted_status_short_url.fillna('', inplace=True)
    df.quoted_status_expand_url.fillna('', inplace=True)
    df.in_reply_to_status_id_str.fillna('', inplace=True)
    df.in_reply_to_user_id_str.fillna('', inplace=True)
    df.hashtags.fillna('', inplace=True)
    df.urls.fillna('', inplace=True)
    df.mentions.fillna('', inplace=True)
    return [create_tweet_from_flat_dict(row) for _, row in df.iterrows()]


def read_tweets_from_json_lines_file(file_path: str) -> List[Tweet]:
    """Method to read tweets from csv file."""
    file = open(file_path, 'r')
    return [create_tweet_from_json(line) for line in file.readlines()]



"""TweetOutput which saves tweet data by request received tweets batch to CSV file."""

import os
from typing import List

import pandas as pd

from .tweet_output import TweetOutput
from ..mapper.tweet_dict_mapper import tweet_to_flat_dict
from ..model.tweet import Tweet

_fields_list = ['created_at', 'id_str', 'conversation_id_str', 'full_text', 'lang', 'favorited', 'retweeted',
                'retweet_count', 'favorite_count', 'reply_count', 'quote_count', 'quoted_status_id_str',
                'quoted_status_short_url', 'quoted_status_expand_url', 'user_id_str', 'user_name', 'user_full_name',
                'user_verified', 'in_reply_to_status_id_str', 'in_reply_to_user_id_str', 'hashtags', 'mentions',
                'urls']


class CsvTweetOutput(TweetOutput):
    """CsvTweetOutput saves tweet data by request received tweets batch to CSV file."""

    file_location: str
    add_header_on_start: bool

    def __init__(self, file_location: str, add_header_on_start: bool = True):
        """Create instance of CsvTweetOutput."""
        self.file_location = file_location
        self.add_header_on_start = add_header_on_start

    def export_tweets(self, tweets: List[Tweet]):
        """Export tweets to CSV."""
        dict_list = [tweet_to_flat_dict(it) for it in tweets]
        df = pd.DataFrame(dict_list, columns=_fields_list)
        df.to_csv(
            path_or_buf=self.file_location,
            mode='a',
            header=self._header_to_add(),
            index=False
        )
        return

    def _header_to_add(self):
        return (not self._file_exists()) or self._file_is_empty()

    def _file_exists(self) -> bool:
        return os.path.isfile(self.file_location)

    def _file_is_empty(self) -> bool:
        return os.stat(self.file_location).st_size == 0



"""Methods to export_data tweets."""
from typing import List

from .util import clear_file
from ..model import Tweet
from ..tweet_output import CsvTweetOutput
from ..tweet_output import JsonLineFileTweetOutput


def export_tweets_to_csv(tweets: List[Tweet], filename: str):
    """Method to export_data tweets to csv."""
    clear_file(filename)
    CsvTweetOutput(filename).export_tweets(tweets)
    return


def export_tweets_to_json_lines(tweets: List[Tweet], filename: str):
    """Method to export_data tweets to json lines file."""
    clear_file(filename)
    JsonLineFileTweetOutput(filename).export_tweets(tweets)
    return



from .replies_filter import RepliesFilter
from .search_run_context import SearchRunContext
from .search_runner import TweetSearchRunner
from .search_tweets_result import SearchTweetsResult
from .search_tweets_task import SearchTweetsTask



from typing import List

import stweet as st


class TweetOutputExportCallCounter(st.TweetOutput):
    counter: int

    def __init__(self):
        self.counter = 0

    def export_tweets(self, tweets: List[st.Tweet]):
        self.counter += 1
        return

    def get_output_call_count(self) -> int:
        return self.counter



"""Tweet - json mapper."""
import json

from arrow import Arrow

from .tweet_dict_mapper import create_tweet_from_dict
from ..model import Tweet


def tweet_to_json(tweet: Tweet) -> str:
    """Method to prepare json of tweet. Used in JSON serialization."""
    return json.dumps(tweet, default=lambda o: str(o) if isinstance(o, Arrow) else o.__dict__)


def create_tweet_from_json(json_value: str) -> Tweet:
    """Method creates tweet from json string."""
    tweet_dict = json.loads(json_value)
    if 'media_url' not in tweet_dict:
        tweet_dict['media_url'] = ''
    return create_tweet_from_dict(tweet_dict)



"""UserOutput which print each n-th tweet."""

from typing import List

from .user_output import UserOutput
from ..model import User


class PrintEveryNUserOutput(UserOutput):
    """UserOutput which print each n-th user."""

    each_n: int
    _counter: int = 0

    def __init__(self, each_n: int):
        """Create instance of PrintEveryNUserOutput."""
        self.each_n = each_n

    def export_users(self, users: List[User]):
        """Method print first each n user in console with all scrapped count."""
        for it in users:
            self._counter += 1
            if self._counter % self.each_n == 0:
                print(self._counter, it)
        return
